using System;
using System.Collections.Generic;
using System.Text;

namespace Sketcher
{
    /*
     * [TODO] Current approach of checking each target for each event will not scale. 
     *   Need to have masks, so we can create short-lists of who to check.
     *   Also should be caching Enabled states (and somehow figuring out out enabled-ness visibility...)
     */


    class InputBehaviorManager
    {
        List<InputBehavior> m_targets;

        InputBehavior m_capturing;
        bool m_bKillCaptureForThisStroke;


        public void Register(InputBehavior behavior) {
            if ( ! m_targets.Contains(behavior) )
                m_targets.Add(behavior);
        }
        public void DeRegister(InputBehavior behavior) {
            m_targets.Remove(behavior);
        }

        public void OnHover(Vector2f v) {
            foreach (InputBehavior r in m_targets) {
                if (r.Enabled)
                    r.OnHover(v);
            }
        }


        public InputBehavior Capturing {
            get { return m_capturing; }
        }


        InputBehavior CheckCapture(Stroke s) {
            float fBestCapture = 0.0f;
            InputBehavior bestCapture = null;
            foreach (InputBehavior r in m_targets) {
                if (!r.Enabled)
                    continue;

                float fCapture = r.GetCaptureScore(s);
                if (fCapture > fBestCapture) {
                    fBestCapture = fCapture;
                    bestCapture = r;
                }
            }
            if (fBestCapture > 0 && fBestCapture != InputBehavior.PREVENT_CAPTURE) {
                return bestCapture;
            } else if (fBestCapture == InputBehavior.PREVENT_CAPTURE)
                m_bKillCaptureForThisStroke = true;
            
            return null;
        }


        public void OnStrokeBegin(Stroke s) {
            m_capturing = null;
            m_bKillCaptureForThisStroke = false;

            InputBehavior capture = CheckCapture(s);
            if (capture != null) {
                bool bOK = capture.OnCaptureStrokeStart(s);
                if ( bOK )
                    m_capturing = capture;
                else
                    System.Diagnostics.Debugger.Break();  // WTF?
            }
        }

        public void OnStrokeUpdate(Stroke s) {
            if (m_capturing != null) {
                bool bContinue = m_capturing.OnCapturedStrokeUpdate(s);
                if (bContinue == false)
                    m_capturing = null;
            } else if (! m_bKillCaptureForThisStroke) {
                InputBehavior capture = CheckCapture(s);
                if (capture != null) {
                    bool bOK = capture.OnCaptureStrokeStart(s);
                    if (bOK)
                        m_capturing = capture;
                    else
                        System.Diagnostics.Debugger.Break();  // WTF?
                }
            }
        }

        public void OnStrokeEnd(Stroke s) {

            // if stroke was captured and consumed, we are done with it
            if ( m_capturing != null ) {
                bool bConsumed = m_capturing.OnCaptureStrokeEnd(s);
                m_capturing = null;
                if ( bConsumed )
                    return;
            }
            
            // do stroke recognition & processing
            float fCircle = s.IsCircleStroke();
            float fLine = s.IsLineStroke();
            float fClick = s.IsClickStroke();
            System.Console.WriteLine("C: " + fCircle + " L: " + fLine + " C: " + fClick);

            bool bUsed = false;

            if (fClick > 0) {
                ClickStroke cls = new ClickStroke(s);
                bUsed = TryClick(cls);
                if (!bUsed) {
                    CircleStroke cs = new CircleStroke(s);
                    bUsed = TryCircle(cs);
                }
            }

            if ( ! bUsed && fCircle > fLine) {
                CircleStroke cs = new CircleStroke(s);
                bUsed = TryCircle(cs);
            } 
            
            if ( ! bUsed && fLine > 0.2f ) {
                LineStroke ls = new LineStroke(s);
                bUsed = TryCross(ls);
            } 

            if ( ! bUsed ) {
                bUsed = TryGeneralStroke(s);
            }

        }


        bool TryCross(LineStroke s) {
            FeatureSet vSet = new FeatureSet();
            for ( int i = 0; i < m_targets.Count; ++i ) {
                if ( ! m_targets[i].Enabled )
                    continue;

                FeatureVector fVec = m_targets[i].GetCrossFeatures(s);
                if ( fVec != null && fVec.Count > 0 ) {
                    fVec.Identifier = i;
                    vSet.Add(fVec);
                }
            }
            FeatureVector fMaxAverage = vSet.MaxAverageValueList;
            if ( fMaxAverage == null )
                return false;
            m_targets[fMaxAverage.Identifier].OnCross(s);
            return true;

                



            //float fMaxScore = 0;
            //InputBehavior rMax = null;
            //foreach (InputBehavior r in m_targets) {
            //    float fScore = r.GetCrossScore(s);
            //    if (fScore > fMaxScore) {
            //        fMaxScore = fScore;
            //        rMax = r;
            //    }
            //}
            //if (rMax == null)
            //    return false;
            //rMax.OnCross(s);
            //return true;
        }


        bool TryCircle(CircleStroke s) {
            float fMaxScore = 0;
            InputBehavior rMax = null;
            foreach (InputBehavior r in m_targets) {
                if ( ! r.Enabled )
                    continue;

                float fScore = r.GetCircleScore(s);
                if (fScore > fMaxScore) {
                    fMaxScore = fScore;
                    rMax = r;
                }
            }
            if (rMax == null)
                return false;
            rMax.OnCircle(s);
            return true;
        }


        bool TryClick(ClickStroke s) {
            float fMaxScore = 0;
            InputBehavior rMax = null;
            foreach (InputBehavior r in m_targets) {
                if (!r.Enabled)
                    continue;

                float fScore = r.GetClickScore(s);
                if (fScore > fMaxScore) {
                    fMaxScore = fScore;
                    rMax = r;
                }
            }
            if (rMax == null)
                return false;
            rMax.OnClick(s);
            return true;
        }


        bool TryGeneralStroke(Stroke s) {
            float fMaxScore = 0;
            InputBehavior rMax = null;
            foreach (InputBehavior r in m_targets) {
                if (!r.Enabled)
                    continue;

                float fScore = r.GetGeneralStrokeScore(s);
                if (fScore > fMaxScore) {
                    fMaxScore = fScore;
                    rMax = r;
                }
            }
            if (rMax == null)
                return false;
            rMax.OnGeneralStroke(s);
            return true;
        }

        static public InputBehaviorManager Get {
            get {
                if (s_singleton == null)
                    s_singleton = new InputBehaviorManager();
                return s_singleton;
            }
        }

        protected InputBehaviorManager() {
            m_targets = new List<InputBehavior>();
            m_capturing = null;
            m_bKillCaptureForThisStroke = false;
        }
        static InputBehaviorManager s_singleton = null;
    }
}
