using System;
using System.Collections.Generic;
using System.Text;

namespace Sketcher
{
    abstract class Container : Item
    {
#region Container interface
        virtual protected void OnItemSetChanged(object o, ItemSet.ItemSetChangedArgs e) {
            if (e.added != null) {
                foreach (Item i in e.added) {
                    i.Parent = this;
                    i.ItemBoundsChangedEvent += m_childItemResizedHandler;
                    i.ElementChangedEvent += m_childItemChangedHandler;
                    i.ElementVisibilityChangedEvent += m_childItemVisibilityHandler;
                }
            } 
            if (e.removed != null) {
                foreach (Item i in e.removed) {
                    i.Parent = null;
                    i.ItemBoundsChangedEvent -= m_childItemResizedHandler;
                    i.ElementChangedEvent -= m_childItemChangedHandler;
                    i.ElementVisibilityChangedEvent -= m_childItemVisibilityHandler;
                }
            }

            m_layout.InvalidateLayout();
            NotifyElementChanged();
        }

        virtual protected void OnChildItemChanged(Element e) {
            NotifyElementChanged();
        }

        virtual protected void OnChildItemVisibilityChanged( Element e ) {
            m_layout.InvalidateLayout();
            NotifyElementChanged();
        }

        virtual protected void OnChildItemResized(Item item) {
            m_layout.InvalidateLayout();
            NotifyElementChanged();
        }

        virtual protected Box2f GetChildAreaRegion() {
            return Bounds;
        }

#endregion


#region Element overrides
        //! disconnect element from the world
        override public void Disconnect( ) {
            base.Disconnect();
            for (int i = 0; i < m_items.Count; ++i)
                m_items[i].Disconnect();
            m_items.Clear();
        }

        override public void Reconnect() {
            // not supported (child connections are broken!!)
            System.Diagnostics.Debugger.Break();
        }

#endregion

        #region Item overrides
        override public void Resize( Box2f newBounds, Transform relativeTo ) {
            base.Resize(newBounds, relativeTo);
            InvalidateLayout();
        }
#endregion



        protected Layout m_layout;
        protected ItemSet m_items;

        // child area information
        Transform m_childAreaOffsetTransform;   //! internal transformation for window boundaries, etc
        Transform m_childAreaScrollTransform;   //! internal transformation for scrolling
        Transform m_childAreaTransform;         //! == Item.Transformation + .OffsetTransform + .ScrollTransform
        Size2f m_childAreaSize;

        ItemBoundsChangedHandler m_childItemResizedHandler;
        ElementChangedHandler m_childItemChangedHandler;
        ElementVisibilityChangedHandler m_childItemVisibilityHandler;

        public Container(Layout layout) {
            m_layout = layout;
            m_items = new ItemSet();

            m_items.ItemSetChangedEvent += new ItemSet.ItemSetChangedHandler(OnItemSetChanged);
            m_childItemChangedHandler = new ElementChangedHandler(OnChildItemChanged);
            m_childItemResizedHandler = new ItemBoundsChangedHandler(OnChildItemResized);
            m_childItemVisibilityHandler = new ElementVisibilityChangedHandler(OnChildItemVisibilityChanged);

            m_childAreaOffsetTransform = new Transform();
            m_childAreaOffsetTransform.PushTransform(Vector2f.ZERO);
            m_childAreaScrollTransform = new Transform();
            m_childAreaScrollTransform.PushTransform(Vector2f.ZERO);
            m_childAreaSize = new Size2f();
            m_childAreaTransform = new Transform();
            m_childAreaTransform.PushTransform(this.Transformation);
            m_childAreaTransform.PushTransform(m_childAreaOffsetTransform);
            m_childAreaTransform.PushTransform(m_childAreaScrollTransform);
        }

        public Layout Layout {
            get { return m_layout; }
            set { m_layout = value; }
        }

        public ItemSet Items {
            get { return m_items; }
        }

        public Size2f ChildAreaSize {
            get { return m_childAreaSize; }
        }
        public Transform ChildAreaScrollTransformation
        {
            get { return m_childAreaScrollTransform; }
        }
        public Transform ChildAreaOffsetTransformation
        {
            get { return m_childAreaOffsetTransform; }
        }
        public Transform ChildAreaTransformation {
            get { return m_childAreaTransform; }
        }

        public Box2f ChildAreaBounds {
            get { return GetChildAreaRegion(); }
        }

        public Vector2f ToChildArea( Vector2f vWindow ) {
            return ChildAreaTransformation.Apply( vWindow, true );
        }

        public void InvalidateLayout() {
            m_layout.InvalidateLayout();
        }

        public void ValidateLayout() {
            if (!m_layout.Valid) {
                Box2f childArea = GetChildAreaRegion();

                //! need to make relative to transformation
                childArea = Transformation.Apply(childArea, true);

                //! update childarea bits
                m_childAreaSize.Width = childArea.Width;
                m_childAreaSize.Height = childArea.Height;

                //[RMS] not sure what this is for...?
                ChildAreaOffsetTransformation.PopTransform();
                ChildAreaOffsetTransformation.PushTransform(childArea.Min);

                m_layout.ValidateLayout(this);
            }
        }


        public void ResizeChild(Item child, Box2f bounds) {
            // [TODO] check if this still works...
            System.Diagnostics.Debugger.Break();
            bounds = ChildAreaTransformation.Apply(bounds, true);
            child.Resize(bounds);
        }



        public void ScrollChildArea(Vector2f vTranslate, bool bRelative) {
            if (bRelative)
                vTranslate += ChildAreaScrollTransformation.Translation;

            if (vTranslate.Y < 0)
                vTranslate.Y = 0;
            else if (vTranslate.Y > m_layout.Bounds( this ).Height - ChildAreaSize.Height) {
                vTranslate.Y = m_layout.Bounds( this ).Height - ChildAreaSize.Height;
            }

            ChildAreaScrollTransformation.PopTransform();
            ChildAreaScrollTransformation.PushTransform( vTranslate );

            NotifyElementChanged();
        }



        public void TranslateChildArea( Vector2f vTranslate, bool bRelative ) {
            if ( bRelative )
                vTranslate += ChildAreaScrollTransformation.Translation;

            ChildAreaScrollTransformation.PopTransform();
            ChildAreaScrollTransformation.PushTransform(vTranslate);

            NotifyElementChanged();
        }


        /*
         *  Utility Functions
         */

        public Box2f MoveRelativeToTop(Box2f box) {
            Box2f bounds = ChildAreaBounds;

            float bottom = 0;
            float yoff = box.Min.Y - bottom;

            float top = bounds.Height;
            float newy = top - yoff - box.Height;

            box.MoveMin(box.Min.X, newy);

            return box;
        }

        public void MoveRelativeToTop(Item item) {
            Box2f bounds = ChildAreaBounds;
            Box2f box = item.Bounds;

            float bottom = bounds.Min.Y;
            float yoff = box.Min.Y - bottom;
            
            float top = bounds.Max.Y;
            float newy = top - yoff - box.Height;

            box.MoveMin(box.Min.X, newy);

            item.Resize(box);
        }

        public void MoveRelativeToTop2(Item item) {
            Box2f bounds = ChildAreaBounds;
            Box2f box = item.Bounds;

            float bottom = 0;
            float yoff = box.Min.Y - bottom;

            float top = bounds.Height;
            float newy = top - yoff - box.Height;

            box.MoveMin(box.Min.X, newy);

            item.Resize(box);
        }

    }
}
