//#define USE_TILES


using System;
using System.Collections.Generic;
using System.Text;
using Tao.OpenGl;


namespace Sketcher
{
    class InfiniteRectContainer_TiledRenderCache : InfiniteRectContainer
    {
        class Tile : RenderCache.RenderCacheTarget {
            InfiniteRectContainer_TiledRenderCache m_container;
            RenderCache m_cache;
            int[] m_nViewport;
            Box2f m_bounds;

            public bool bDirty;

            public Tile( InfiniteRectContainer_TiledRenderCache container, int[] viewport ) {
                if (viewport[2] != viewport[3])
                    throw new ArgumentException( "Invalid viewport - must be square and power-of-2!" );

                m_container = container;
                m_nViewport = new int[4];
                viewport.CopyTo( m_nViewport, 0 );
                m_cache = new RenderCache( this );
                m_bounds = new Box2f( m_nViewport[0], m_nViewport[1], m_nViewport[0]+m_nViewport[2], m_nViewport[1]+m_nViewport[3] );
            }

            public Box2f Bounds { 
                get { return m_bounds; }
            }

            public int[] Viewport {
                get { return m_nViewport; }
            }
            public int Width {
                get { return m_nViewport[2]; }
            }
            public int Height {
                get { return m_nViewport[3]; }
            }


            public void Update( ) {
                m_cache.Invalidate();
                m_cache.Validate();
            }

            public void RenderCacheableContent( ) {
                m_container.Render_Tile( this );
            }

            public void RenderTile( ) {
                RenderEngine.Get.PushTransform();
                RenderEngine.Get.Translate( Bounds.Min );
                m_cache.RestoreFromCache();
                RenderEngine.Get.PopTransform();
            }
        };

        //! must be power-of-two
        int m_nTileSize;

        //! sorted first by row, then by column
        SortedList<int, SortedList<int, Tile>> m_tiles;

        List<Element> m_newElements;
        List<Element> m_dirtyElements;
        List<Tile> m_dirtyTiles;
        Box2f m_dirty;

        public InfiniteRectContainer_TiledRenderCache( Box2f bounds ) : base(bounds, new FreeLayout())
        {
            m_tiles = new SortedList<int, SortedList<int, Tile>>();
            m_nTileSize = 128;
            m_dirty = Box2f.INFINITE;
            m_dirtyElements = new List<Element>();
            m_newElements = new List<Element>();
            m_dirtyTiles = new List<Tile>();
        }



        protected override void OnItemSetChanged( object o, ItemSet.ItemSetChangedArgs e ) {
            base.OnItemSetChanged( o, e );
#if USE_TILES
            if (e.added != null) {
                foreach (Item i in e.added)
                    m_newElements.Add(i);
            }
#endif
        }

        protected override void OnChildItemChanged( Element e ) {
            base.OnChildItemChanged( e );
#if USE_TILES
            m_dirtyElements.Add( e );
#endif
        }
        protected override void OnChildItemResized( Item item ) {
            base.OnChildItemResized( item );
#if USE_TILES
            m_dirtyElements.Add( item );
#endif
        }




        void ToTileIndexBounds( Box2f bounds, out int nColMin, out int nColMax, out int nRowMin, out int nRowMax ) {
            // convert bounds to integers
            int nLowX = (int)Math.Floor( bounds.Min.X );
            int nHighX = (int)Math.Ceiling( bounds.Max.X );
            int nLowY = (int)Math.Floor( bounds.Min.Y );
            int nHighY = (int)Math.Ceiling( bounds.Max.Y );

            nColMin = (nLowX >= 0) ? nLowX / m_nTileSize : (nLowX / m_nTileSize) - 1;
            nColMax = (nHighX >= 0) ? nHighX / m_nTileSize : (nHighX / m_nTileSize) - 1;
            nRowMin = (nLowY >= 0) ? nLowY / m_nTileSize : (nLowY / m_nTileSize) - 1;
            nRowMax = (nHighY >= 0) ? nHighY / m_nTileSize : (nHighY / m_nTileSize) - 1;
        }


        void AddTiles(Box2f bounds) {
            int nRowStart, nColStart, nRowEnd, nColEnd;
            ToTileIndexBounds( bounds, out nColStart, out nColEnd, out nRowStart, out nRowEnd );
            for (int r = nRowStart; r <= nRowEnd; ++r) {
                if (!m_tiles.ContainsKey( r ))
                    m_tiles.Add( r, new SortedList<int, Tile>() );
                SortedList<int, Tile> rowTiles = m_tiles[r];

                for (int c = nColStart; c <= nColEnd; ++c) {
                    Tile t = FindTile( r, c );
                    if (t == null) {
                        int[] nViewport = new int[4];
                        nViewport[0] = c * m_nTileSize;
                        nViewport[1] = r * m_nTileSize;
                        nViewport[2] = m_nTileSize;
                        nViewport[3] = m_nTileSize;
                        t = new Tile( this, nViewport );
                        rowTiles.Add( c, t );
                    }
                }
            }
        }


        Tile FindTile( int row, int col ) {
            if (!m_tiles.ContainsKey( row ))
                return null;
            SortedList<int, Tile> rowTiles = m_tiles[row];
            if (!rowTiles.ContainsKey( col ))
                return null;
            return rowTiles[col];
        }


        void ValidateTiles( ) {

            // [HACK] clear dirty flags on each tile
            foreach ( int r in m_tiles.Keys )
                foreach ( int c in m_tiles[r].Keys )
                    m_tiles[r][c].bDirty = false;

            // create any necessary new tiles
            if (m_newElements.Count > 0) {
                foreach (Element e in m_newElements) {
                    m_dirtyElements.Add( e );

                    Box2f childBounds = ChildAreaTransformation.Apply( e.Bounds, true );
                    AddTiles( childBounds );
                }
                m_newElements.Clear();
            }


            if (m_dirtyElements.Count == 0)
                return;

            // figure out dirty rectangle (or could manually tag dirty tiles instead....?)
            m_dirty = m_dirtyElements[0].Bounds;
            for (int i = 1; i < m_dirtyElements.Count; ++i)
                m_dirty.Contain( m_dirtyElements[i].Bounds );
            m_dirty = ChildAreaTransformation.Apply( m_dirty, true );

            m_dirtyTiles.Clear();

            int nRowStart, nColStart, nRowEnd, nColEnd;
            ToTileIndexBounds( m_dirty, out nColStart, out nColEnd, out nRowStart, out nRowEnd );
            for (int r = nRowStart; r <= nRowEnd; ++r) {
                for (int c = nColStart; c <= nColEnd; ++c) {
                    Tile t = FindTile( r, c );
                    if (t != null)
                        m_dirtyTiles.Add( t );
                }
            }


            // update dirty tiles
            foreach (Tile t in m_dirtyTiles) {
                t.bDirty = true;
                t.Update();
            }


            m_dirtyElements.Clear();
            m_dirty = Box2f.EMPTY;
        }







        void Render_Tile( Tile t ) {

            Box2f tileBounds = t.Bounds;
            Box2f absoluteTileBounds = ChildAreaTransformation.Apply( tileBounds );
            RenderEngine.Get.PushDisableClipBox();

            Gl.glMatrixMode( Gl.GL_PROJECTION );
            Gl.glPushMatrix();
            Gl.glLoadIdentity();
            Gl.glOrtho( 0, m_nTileSize, 0, m_nTileSize, -1, 1 );
            Gl.glMatrixMode( Gl.GL_MODELVIEW );

            Gl.glClear( Gl.GL_COLOR_BUFFER_BIT );

            RenderEngine.Get.PushClearTransform();
            RenderEngine.Get.Translate( -1.0f * tileBounds.Min );

            for (int i = 0; i < Items.Count; ++i) {
                if (Element.VisibleForRendering( Items[i].Visibility ) && 
                     Items[i].Bounds.Intersects( absoluteTileBounds ))
                    Items[i].RenderBackground();
            }

            for (int i = 0; i < Items.Count; ++i) {
                if (Element.VisibleForRendering( Items[i].Visibility ) && 
                     Items[i].Bounds.Intersects( absoluteTileBounds ))
                    Items[i].Render();
            }

            RenderEngine.Get.PopTransform();
            RenderEngine.Get.PopClipBox();


            Gl.glMatrixMode( Gl.GL_PROJECTION );
            Gl.glPopMatrix();
            Gl.glMatrixMode( Gl.GL_MODELVIEW );
        }





        override public void Render( ) {
            ValidateLayout();
#if USE_TILES
            ValidateTiles();
#endif

            RenderEngine.Get.DrawBox( Bounds, AlphaBackgroundColor );
            RenderEngine.Get.DrawBox( ChildAreaBounds, AlphaForegroundColor );

            // set clip region and scroll transformation for child area
            RenderEngine.Get.PushTransform( ChildAreaTransformation );
            RenderEngine.Get.PushClipBox( ChildAreaBounds );

#if USE_TILES
            RenderEngine.Get.DisableAlphaBlending();
            foreach (int r in m_tiles.Keys)
                foreach (int c in m_tiles[r].Keys)
                    m_tiles[r][c].RenderTile();
            RenderEngine.Get.EnableAlphaBlending();

            // debug - draw tile borders w/ color determining whether tile was dirty or not
            foreach (int r in m_tiles.Keys) {
                foreach (int c in m_tiles[r].Keys) {
                    Tile t =  m_tiles[r][c];
                    RenderEngine.Get.DrawBoxOutline( t.Bounds,
                        t.bDirty ? ColorRGBA.GREEN : ColorRGBA.RED );
                }
            }
#else
            for (int i = 0; i < Items.Count; ++i) {
                if (Element.VisibleForRendering(Items[i].Visibility))
                    Items[i].RenderBackground();
            }
            for (int i = 0; i < Items.Count; ++i) {
                if (Element.VisibleForRendering(Items[i].Visibility))
                    Items[i].Render();
            }    
#endif

            RenderEngine.Get.PopClipBox();
            RenderEngine.Get.PopTransform();
        }

    }
}
