using System;
using System.Collections.Generic;
using System.Text;

namespace Sketcher
{
    class Transform
    {
        enum EntryType { ConstVector, Transform };

        interface Entry
        {
            EntryType Type { get; }
            Vector2f Translation { get; }
        }
        class TransformEntry : Transform.Entry
        {
            public Transform m_Transform;
            public Transform.TransformChangedHandler m_Handler;
            public TransformEntry(Transform vTransform) {
                m_Transform = vTransform;
                m_Handler = null;
            }
            public EntryType Type {
                get { return EntryType.Transform; }
            }
            public Vector2f Translation {
                get { return m_Transform.Translation; }
            }
        }
        class ConstVectorEntry : Transform.Entry
        {
            public Vector2f m_vTranslate;
            public ConstVectorEntry(Vector2f v) {
                m_vTranslate = v;
            }
            public Vector2f Vector {
                get { return m_vTranslate; }
                set { m_vTranslate = value; }
            }
            public EntryType Type {
                get { return EntryType.ConstVector; }
            }
            public Vector2f Translation {
                get { return m_vTranslate; }
            }
        }

        Stack<Entry> m_vTransformStack;
        Stack<Vector2f> m_vTranslateStack;

        public Transform() {
            m_vTransformStack = new Stack<Entry>();
            m_vTranslateStack = new Stack<Vector2f>();
        }

        public Vector2f Translation {
            get {
                if (m_vTranslateStack.Count == 0)
                    return Vector2f.ZERO;
                else
                    return m_vTranslateStack.Peek();
            }
        }

        public void Clear() {
            m_vTranslateStack.Clear();
            m_vTransformStack.Clear();

            NotifyTransformChanged();
        }


        public void PushTransform(Vector2f vTranslate) {
            m_vTransformStack.Push(new ConstVectorEntry(vTranslate));
            m_vTranslateStack.Push(Translation + vTranslate);

            NotifyTransformChanged();
        }

        public void PushTransform(Transform t) {
            TransformEntry entry = new TransformEntry(t);
            m_vTransformStack.Push(entry);
            m_vTranslateStack.Push(Translation + t.Translation);

            entry.m_Handler = new Transform.TransformChangedHandler(ParentTransformChangedHandler);
            t.TransformChangedEvent +=  entry.m_Handler;

            NotifyTransformChanged();
        }

        public void PopTransform() {
            if ( m_vTranslateStack.Count > 0 )
                m_vTranslateStack.Pop();
            if (m_vTransformStack.Count > 0) {
                Entry e = m_vTransformStack.Pop();
                if (e.Type == EntryType.Transform) {
                    TransformEntry entry = (TransformEntry)e;
                    entry.m_Transform.TransformChangedEvent -= entry.m_Handler;
                }
            }

            NotifyTransformChanged();
        }

        //! this function assumes that the bottom of the stack is a ConstVectorEntry, and just forces it's value
        public void SetTransform(Vector2f vTranslate) {
            if (m_vTransformStack.Count == 0 || m_vTransformStack.Peek().Type != EntryType.ConstVector)
                throw new Exception("Error in SetTransform!");
            ((ConstVectorEntry)m_vTransformStack.Peek()).Vector = vTranslate;
            m_vTranslateStack.Pop();
            m_vTranslateStack.Push(vTranslate);

            NotifyTransformChanged();
        }


        public Vector2f Apply(Vector2f vSrc) {
            return vSrc + Translation;
        }
        public Box2f Apply(Box2f vSrc) {
            vSrc.Translate(Translation);
            return vSrc;
        }
        public Vector2f Apply(Vector2f vSrc, bool bInverse) {
            return (bInverse) ? vSrc - Translation : vSrc + Translation;
        }
        public Box2f Apply(Box2f vSrc, bool bInverse) {
            vSrc.Translate((bInverse) ? Translation*-1.0f : Translation);
            return vSrc;
        }




        // notification handler for item changes
        public delegate void TransformChangedHandler(object o);
        public event TransformChangedHandler TransformChangedEvent;

        protected void NotifyTransformChanged() {
            if (TransformChangedEvent != null)
                TransformChangedEvent(this);
        }

        // handler for when parent transforms change
        public void ParentTransformChangedHandler(object o) {
            // RMS TODO: instead of rebuilding here, set invalid flag, and check before returning

            RebuildStack();
        }


        void RebuildStack() {
            m_vTranslateStack.Clear();

            List<Vector2f> vTranslations = new List<Vector2f>();
            foreach (Entry entry in m_vTransformStack)
                vTranslations.Add(entry.Translation);
            vTranslations.Reverse();

            Vector2f vCur = Vector2f.ZERO;
            foreach (Vector2f vTranslate in vTranslations) {
                vCur += vTranslate;
                m_vTranslateStack.Push(vCur);
            }

            NotifyTransformChanged();
        }
    }
}
