using System;
using System.Collections.Generic;
using System.Text;
using Tao.OpenGl;
using Tao.Platform.Windows;

namespace Sketcher
{
    class GLCanvas : RenderCache.RenderCacheTarget
    {
        OpenGLWinFormsControl m_glcontrol;

        Container m_toplevel;
        FitToViewportContainer m_topLevelFit;

        public delegate string ApplicationTitleString();
        ApplicationTitleString m_appTitle;

        int[] m_nCurrentViewport = {0,0,0,0};

        RenderCache m_overlay;

        bool m_bFullPaintPending;
        bool m_bOverlayPaintPending;

        public GLCanvas(OpenGLWinFormsControl control) {
            m_glcontrol = control;

            m_appTitle = null;

            m_toplevel = null;
            m_topLevelFit = null;

            WidgetManager.Get.WidgetRedrawEvent += new WidgetManager.WidgetRedrawHandler(WidgetChangedHandler);

            // start animation
            AnimationManager.Get.Start();

            m_overlay = new RenderCache(this);
        }


        public int[] Viewport {
            get { return m_nCurrentViewport; }
        }
        public int Width {
            get { return m_nCurrentViewport[2]; }
        }
        public int Height {
            get { return m_nCurrentViewport[3]; }
        }

        public String ApplicationTitle {
            get { return (m_appTitle == null) ? "GLCanvas" : m_appTitle(); }
        }

        public RenderCache Overlay {
            get { return m_overlay; }
        }


        /*
         * rendering
         */

        public void Initialize() {
            Gl.glGetIntegerv(Gl.GL_VIEWPORT, m_nCurrentViewport);

            Gl.glMatrixMode(Gl.GL_PROJECTION);
            Gl.glLoadIdentity();
            Gl.glOrtho(0, m_nCurrentViewport[2], 0, m_nCurrentViewport[3], -1, 1);
            Gl.glMatrixMode(Gl.GL_MODELVIEW);
            Gl.glLoadIdentity();

            m_overlay.Initialize();

            // pre-load textures, create static display lists, etc
            RenderEngine.Get.Initialize();

            m_bFullPaintPending = true;
        }


        public void Resized(int nWidth, int nHeight) {
            Gl.glGetIntegerv(Gl.GL_VIEWPORT, m_nCurrentViewport);

            Gl.glMatrixMode(Gl.GL_PROJECTION);
            Gl.glLoadIdentity();
            Gl.glOrtho(0, m_nCurrentViewport[2], 0, m_nCurrentViewport[3], -1, 1);
            Gl.glMatrixMode(Gl.GL_MODELVIEW);
            Gl.glLoadIdentity();

            if (m_topLevelFit != null)
                m_topLevelFit.ViewportUpdated();

            Overlay.Invalidate();

            m_bFullPaintPending = true;
        }



        public void Render() {
            //System.Diagnostics.Debug.WriteLine("[Render]");

            // [TODO] handle this a better way...
            // This case happens when windowing system asks us to redraw. Could use overlay background
            //  instead of drawing scene, if we had it available...
            if (!m_bFullPaintPending && !m_bOverlayPaintPending) {
                Render_FullPaint();
                return;
            }


            bool bDidFullPaint = false;
            if (m_bFullPaintPending) {
                Render_FullPaint();
                Overlay.Invalidate();
                m_bFullPaintPending = false;
                bDidFullPaint = true;
                //System.Diagnostics.Debug.WriteLine("  [FullPaint]");
            }

            if (m_bOverlayPaintPending) {
                Render_Overlay(bDidFullPaint);
                m_bOverlayPaintPending = false;
                //System.Diagnostics.Debug.WriteLine("  [Overlay]");
            }
        }


        public void Render_Overlay(bool bBackBufferAlreadyValid) {
            if ( ! bBackBufferAlreadyValid ) {
                Overlay.Validate();
                Overlay.RestoreFromCache();
            }

            RenderingBehaviorManager.Get.OverlayRender(null);
        }


        public void Render_FullPaint() {
            Gl.glClearColor(1, 1, 1, 1.0f);
            Gl.glClear(Gl.GL_DEPTH_BUFFER_BIT | Gl.GL_COLOR_BUFFER_BIT);

            RenderEngine.Get.EnableAlphaBlending();

            m_toplevel.Render();
            WidgetManager.Get.Render();

            RenderEngine.Get.DisableAlphaBlending();
        }




        public void PostOverlayRepaint() {
            m_bOverlayPaintPending = true;
            m_glcontrol.Invalidate();
        }


        public void PostRepaint() {
            m_bFullPaintPending = true;
            Overlay.Invalidate();
            m_glcontrol.Invalidate();
        }

        public void AnimationTick() {
            if ( AnimationManager.Get.AnimationTick() )
                PostRepaint();
        }


        /*
         * input
         */

        public void OnHover(Vector2f v) {
            InputBehaviorManager.Get.OnHover(v);
            if (RenderingBehaviorManager.Get.OverlayRendererRequested(null))
                PostOverlayRepaint();
        }
        public void OnStrokeBegin(Stroke s) {
            InputBehaviorManager.Get.OnStrokeBegin(s);

            object captureItem = InputBehaviorManager.Get.Capturing;
            if ( captureItem != null ) {
                if ( RenderingBehaviorManager.Get.OverlayRendererRequested( captureItem ) )
                    PostOverlayRepaint();
            }
        }
        public void OnStrokeUpdate(Stroke s) {
            InputBehaviorManager.Get.OnStrokeUpdate(s);

            object captureItem = InputBehaviorManager.Get.Capturing;
            if (captureItem != null) {
                if (RenderingBehaviorManager.Get.OverlayRendererRequested(captureItem))
                    PostOverlayRepaint();
            }
        }
        public void OnStrokeEnd(Stroke s) {
            InputBehaviorManager.Get.OnStrokeEnd(s);
        }




        public void InitializeTest() {
            //InitializeCrosskeyTest();
            //InitializeFingerKeyTest();

            InitializeVectorCanvasTest();

            //InitializeFolderTest();
        }


        public void InitializeCrosskeyTest() {
            int nSize = 400;
            WindowContainer window = new WindowContainer(new Box2f(0, 0, nSize, nSize), "CrossKeyboard");
            window.ElementChangedEvent += new Item.ElementChangedHandler(ItemChangedHandler);
            window.ChildAreaBorder = 0.0f;
            m_toplevel = window;
            
            CrossKeys.CrossKeyboard keyboard = new CrossKeys.CrossKeyboard(new Box2f(0, 0, nSize - 30, nSize - 30));
            m_toplevel.Items.Add(keyboard);
        }

        public void InitializeFingerKeyTest() {
            int nWidth = 600;
            int nHeight = 320;
            WindowContainer window = new WindowContainer(new Box2f(0, 0, nWidth, nHeight), "FingerKeyboard");
            window.ElementChangedEvent += new Item.ElementChangedHandler(ItemChangedHandler);
            window.ChildAreaBorder = 0.0f;
            m_toplevel = window;
            
            CrossKeys.FingerKeyboard keyboard = 
                new CrossKeys.FingerKeyboard(new Box2f(0, 0, nWidth, nHeight));
            m_toplevel.Items.Add(keyboard);
        }


        public void InitializeFolderTest() {
            m_toplevel = new WindowContainer(new Box2f(20, 20, 800, 600), "Test Window");
            m_toplevel.ElementChangedEvent += new Item.ElementChangedHandler(ItemChangedHandler);


            InfiniteRectContainer child = new InfiniteRectContainer(
                new Box2f(0, 0, 100, 100), new VerticalTilingLayout());
            child.Behaviors.Add(new VerticalDragTranslateBehavior(child));

            List<Item> items = new List<Item>();
            for (int i = 0; i < 100; ++i) {
                IconItem si = new IconItem(
                    new Size2f(80, 80), ColorRGBA.RED);
                items.Add(si);
            }
            child.Items.Add(items);

            m_toplevel.Items.Add(child);
        }


        public void InitializeVectorCanvasTest() {
            m_topLevelFit = new FitToViewportContainer();
            m_toplevel = m_topLevelFit;
            
            Notebook notebook = new Notebook(new Box2f(0, 0, 1, 1));
            notebook.ElementChangedEvent += new Item.ElementChangedHandler(ItemChangedHandler);

            //m_appTitle = delegate() { return "Sketcher - " + notebook.NotebookTitleString(); };
            m_appTitle = delegate( ) { return notebook.NotebookTitleString(); };

            m_toplevel.Items.Add(notebook);
        }


        // handler for when items change
        public void ItemChangedHandler(object o) {
            Overlay.Invalidate();
            PostRepaint();
        }

        public void WidgetChangedHandler(Widget w) {
            Overlay.Invalidate();
            PostRepaint();
        }



        public static bool CheckGLErrors() {
            int nError = 0;
            bool bFoundErrors = false;
            while ( (nError = Gl.glGetError() ) != Gl.GL_NO_ERROR ) {
                System.Diagnostics.Debug.WriteLine("Got GL error " + nError );
                bFoundErrors = true;
            }
            return bFoundErrors;
        }



#region RenderTarget Members
        public void RenderCacheableContent( ) {
            Render_FullPaint();
        }
#endregion
    }


}
