using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Drawing;
using System.Drawing.Imaging;

namespace Sketcher
{
    public unsafe class Image
    {
        public struct Pixel {
            public byte r;
            public byte g;
            public byte b;
            public byte a;
            public Pixel(byte red, byte green, byte blue) { r = red; g = green; b = blue; a = 255; }
            public Pixel(byte red, byte green, byte blue, byte alpha) { r = red; g = green; b = blue; a = alpha; }
            public Pixel(byte alpha) { r = 0; g = 0; b = 0; a = alpha; }

            public Pixel(float red, float green, float blue) {
                r = (byte)Math.Max(0, Math.Min(255, red * 255.0f) );
                g = (byte)Math.Max(0, Math.Min(255, green * 255.0f) );
                b = (byte)Math.Max(0, Math.Min(255, blue * 255.0f) ); 
                a = 255;
            }
            public Pixel(float red, float green, float blue, float alpha) {
                r = (byte)Math.Max(0, Math.Min(255, red * 255.0f) );
                g = (byte)Math.Max(0, Math.Min(255, green * 255.0f) );
                b = (byte)Math.Max(0, Math.Min(255, blue * 255.0f) ); 
                a = (byte)Math.Max(0, Math.Min(255, alpha * 255.0f) ); 
            }
            public Pixel(ColorRGBA color) : this(color.R, color.G, color.B, color.A ) {}
            public float fA {
                get { return (float)a / 255.0f; }
                set { a = (byte)Math.Max(0, Math.Min(255, value * 255.0f) ); }
            }

            public static bool operator ==(Pixel p1, Pixel p2) 
                { return (p1.r==p2.r) && (p1.b==p2.b) && (p1.g==p2.g) && (p1.a==p2.a); }
            public static bool operator !=(Pixel p1, Pixel p2) 
                { return (p1.r!=p2.r) || (p1.b!=p2.b) || (p1.g!=p2.g) || (p1.a!=p2.a); }
            public override bool Equals(object obj) 
                { return this == (Pixel)obj; }
            public override int  GetHashCode()
                { return (int)r; }                  // does this make sense?
    
        }

        byte[] m_vBuffer;
        uint m_nWidth;
        uint m_nHeight;
        uint m_nChannels;

        public Image() {
            m_nWidth = (uint)1;
            m_nHeight = (uint)1;
            m_nChannels = 4;
            Resize();
        }
        public Image(int nWidth, int nHeight) {
            m_nWidth = (uint)nWidth;
            m_nHeight = (uint)nHeight;
            m_nChannels = 4;
            Resize();
        }
        public Image(int nWidth, int nHeight, uint nChannels) {
            m_nWidth = (uint)nWidth;
            m_nHeight = (uint)nHeight;
            m_nChannels = nChannels;
            Resize();
        }
        public Image(int nWidth, int nHeight, uint nChannels, byte[] buffer) {
            m_nWidth = (uint)nWidth;
            m_nHeight = (uint)nHeight;
            m_nChannels = nChannels;
            Resize();
            buffer.CopyTo( m_vBuffer, 0 );
        }

        public int Channels {
            get { return (int)m_nChannels; }
            set { m_nChannels = (uint)value; Resize(); }
        }
        public int Width {
            get { return (int)m_nWidth; }
            set { m_nWidth = (uint)value; Resize(); }
        }
        public int Height {
            get { return (int)m_nHeight; }
            set { m_nHeight = (uint)value; Resize(); }
        }
        public void Resize(int nWidth, int nHeight) {
            m_nWidth = (uint)nWidth; m_nHeight = (uint)nHeight; Resize();
        }


        public byte[] Buffer {
            get { return m_vBuffer; }
        }

        public Pixel GetPixel(int x, int y) {
            fixed (byte * pPixel = &m_vBuffer[(uint)y * m_nWidth * m_nChannels + (uint)x * m_nChannels]) {
                return *(Pixel *)pPixel;
            }
        }

        public void SetPixel(int x, int y, Pixel pixel) {
            fixed (byte * pCurPixel = &m_vBuffer[(uint)y * m_nWidth * m_nChannels + (uint)x * m_nChannels]) {
                byte * pSetPixel = &(pixel.r);
                for ( uint i = 0; i < m_nChannels; ++i )
                    pCurPixel[i] = pSetPixel[i];
            }
        }


        public bool CopyBitmap( Bitmap bmp ) {
            UnsafeBitmap ubmp = new UnsafeBitmap( bmp, true );
            Resize( ubmp.Bitmap.Width, ubmp.Bitmap.Height );
            ubmp.LockBitmap();
            for (int yi = 0; yi < m_nHeight; ++yi) {
                for (int xi = 0; xi < m_nWidth; ++xi) {
                    UnsafeBitmap.PixelData pix2 = ubmp.GetPixel( xi, yi );
                    Pixel pix = new Pixel( pix2.red, pix2.green, pix2.blue, pix2.alpha );
                    SetPixel( xi, (int)m_nHeight - 1 - yi, pix );
                }
            }
            ubmp.UnlockBitmap();
            return true;
        }

        public bool Load(string sFilename) {
            Bitmap bmp = null;
            try { 
                bmp = new Bitmap(sFilename);
            } catch ( Exception ) {
                return false;
            }
            return CopyBitmap( bmp );
        }


        public bool Load(FilePath path) {
            Bitmap bmp = null;
            try {
                if (path.Location == FilePath.FileLocation.Resource) {
                    Assembly a = Assembly.GetExecutingAssembly();
                    bmp = new Bitmap(a.GetManifestResourceStream(path.FullName));
                } else
                    bmp = new Bitmap(path.FullName);
            } catch (Exception) {
                return false;
            }
            return CopyBitmap(bmp);
        }


        public bool Save(string sFilename) {
            UnsafeBitmap ubmp = new UnsafeBitmap((int)m_nWidth, (int)m_nHeight, PixelFormat.Format32bppArgb );
            ubmp.LockBitmap();
            for ( int yi = 0; yi < m_nHeight; ++yi ) {
                for ( int xi = 0; xi < m_nWidth; ++xi ) {
                    Pixel pix = GetPixel(xi,yi);
                    UnsafeBitmap.PixelData pix2;
                    pix2.red = pix.r;     pix2.blue = pix.b;
                    pix2.green = pix.g;   pix2.alpha = pix.a;
                    ubmp.SetPixel(xi, (int)m_nHeight - 1 - yi, pix2);
                }
            }
            ubmp.UnlockBitmap();
            try {
                ubmp.Bitmap.Save(sFilename);
                return true;
            } catch (Exception) {
                return false;
            }
        }



        public void Clear(ColorRGBA bgColor) {
            Pixel pix = new Pixel(bgColor);
            for ( int yi = 0; yi < m_nHeight; ++yi )
                for ( int xi = 0; xi < m_nWidth; ++xi )
                    SetPixel(xi, yi, pix);       
        }



        public void Crop(ColorRGBA bgColor, int nBorder, bool bSquare) {
            Pixel cropPix = new Pixel(bgColor);

            int xMin = Width, xMax = 0, yMin = Height, yMax = 0;
            for ( int yi = 0; yi < m_nHeight; ++yi ) {
                for ( int xi = 0; xi < m_nWidth; ++xi ) {
                    Pixel p = GetPixel(xi,yi);
                    if ( p != cropPix ) {
                        xMin = (xi < xMin) ? xi : xMin;
                        xMax = (xi > xMax) ? xi : xMax;
                        yMin = (yi < yMin) ? yi : yMin;
                        yMax = (yi > yMax) ? yi : yMax;
                    }
                }
            }
            ++xMax;  ++yMax;   // want to use <, not <=

            int nNewWidth = xMax - xMin + 2*nBorder,  nNewHeight = yMax - yMin + 2*nBorder;
            int xOff = nBorder, yOff = nBorder;
            if ( bSquare ) {
                if ( nNewWidth > nNewHeight ) {
                    yOff = (nNewWidth-nNewHeight)/2 + nBorder;
                    nNewHeight = nNewWidth;
                } else {
                    xOff = (nNewHeight-nNewWidth)/2 + nBorder;
                    nNewWidth = nNewHeight;
                }
            }


            byte[] oldBuffer = m_vBuffer;
            int nOldWidth = Width, nOldHeight = Height, nOldChannels = (int)Channels;
            Resize(nNewWidth, nNewHeight);
            Clear(bgColor);

            for ( int yi = yMin; yi < yMax; ++yi ) {
                for ( int xi = xMin; xi < xMax; ++xi ) {
                    fixed (byte * pOldPixel = &oldBuffer[yi * nOldWidth * nOldChannels + xi * nOldChannels]) {
                        SetPixel( xi-xMin+xOff, yi-yMin+yOff, new Pixel(pOldPixel[0], pOldPixel[1], pOldPixel[2], pOldPixel[3]) );
                    }
                }
            }
        }



        void Resize() {
            m_vBuffer = new byte[m_nWidth * m_nHeight * m_nChannels];
        }
    }
}
