using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;

namespace Sketcher 
{
    class Notebook : FixedRectContainer
    {
        int m_nCurrentPage;
        FilePath m_path;
        bool m_bDirty;      //! set if file has been modified and needs save

        public Notebook(Box2f bounds) :  base(bounds, new VerticalStackingExpanderLayout(bounds)) {
            m_nCurrentPage = -1;
            AppendPage(true);
            m_path = FilePath.UnknownPath;
            m_bDirty = true;

            ChildAreaBorder = 0;
        }

        public FilePath Path {
            get { return m_path; }
            set { m_path = value;  }
        }

        public bool NotebookModified {
            get { return m_bDirty; }
        }

        public string NotebookTitleString( ) {
            if (m_path == FilePath.UnknownPath)
                return (m_bDirty) ? "(New Notebook)*" : "(New Notebook)";
            else
                return (m_bDirty) ? Path.Filename + "*" : Path.Filename;
        }



        public NotebookPage CurrentPage {
            get { return Items[m_nCurrentPage] as NotebookPage; }
        }

        public void PreviousPage() {
            int nNewPage = Math.Max(m_nCurrentPage-1, 0);
            if ( nNewPage != m_nCurrentPage ) {
                CurrentPage.Hide();
                m_nCurrentPage = nNewPage;
                CurrentPage.Show();
            }
        }
        public void NextPage() {
            int nNewPage = Math.Min(m_nCurrentPage+1, Items.Count-1);
            if (nNewPage != m_nCurrentPage) {
                CurrentPage.Hide();
                m_nCurrentPage = nNewPage;
                CurrentPage.Show();
            }
        }


        public NotebookPage AppendPage(bool bSwitchToPage) {
            if (m_nCurrentPage != -1 && bSwitchToPage)
                CurrentPage.Hide();

            NotebookPage newPage = new NotebookPage(Bounds, this);
            newPage.Visibility = VisibilityState.Hidden;
            Items.Add(newPage);
            newPage.CanvasContentsModifiedEvent += new VectorCanvas.CanvasContentsModifiedHandler( OnCanvasChanged );
            m_bDirty = true;

            if (m_nCurrentPage == -1 || bSwitchToPage) {
                m_nCurrentPage = Items.Count-1;
                if (m_nCurrentPage == 0)
                    newPage.Visibility = VisibilityState.Visible;
                else
                    CurrentPage.Show();
            }

            return newPage;
        }


        //! discard all current pages and re-initialize with first blank page
        public void ClearNotebook( ) {
            while (Items.Count > 0) {
                Item i = Items[0];
                Items.Remove( i );
                i.Disconnect();
            }
            m_nCurrentPage = -1;
            AppendPage( true );
        }



        public void SelectTool(string sToolName) {
            for (int i = 0; i < Items.Count; ++i) {
                NotebookPage page = Items[i] as NotebookPage;
                page.SelectTool(sToolName);
            }
        }



        /*
         *  save/load support
         */

        virtual protected void OnCanvasChanged( VectorCanvas canvas ) {
            m_bDirty = true;
        }

        virtual public void NotifyCanvasStored( ) {
            m_bDirty = false;
        }

        public bool Store(XmlDocument doc, XmlElement notebook_node) {
            for ( int i = 0; i < Items.Count; ++i ) {
                XmlElement page_node = doc.CreateElement("Page");
                notebook_node.AppendChild(page_node);

                NotebookPage page = Items[i] as NotebookPage;
                if (page != null)
                    page.Store(doc, page_node);
            }
            return true;
        }


        public bool Restore( XmlDocument doc, XmlElement notebook_node ) {
            ClearNotebook();

            XmlNodeList vPages = notebook_node.GetElementsByTagName( "Page" );
            for ( int i = 0; i < vPages.Count; ++i ) {
                XmlElement page_node = vPages[i] as XmlElement;
                if (page_node == null)
                    continue;

                NotebookPage page = null;
                if (i == 0)
                    page = Items[0] as NotebookPage;
                else
                    page = AppendPage( false );

                page.Restore( doc, page_node );

            }

            m_bDirty = false;

            return true;
        }


    }
}
