using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;

namespace Sketcher
{
    class VectorCanvas : InfiniteRectContainer_TiledRenderCache
    {
        protected Selection m_selection;
        protected VectorToolSet m_tools;

        List<VectorGraphicsItem> m_vPendingOptimize;

        UndoStack m_undoStack;

        public VectorCanvas(Box2f bounds) : base(bounds) 
        {
            m_tools = new VectorToolSet(this);
            m_tools.AddTool( new VectorStrokeTool(this) );
            m_tools.AddTool(new VectorEraserTool(this));
            SelectTool(VectorStrokeTool.ToolName);

            m_vPendingOptimize = new List<VectorGraphicsItem>();
            m_undoStack = new UndoStack();

            m_selection = new Selection();
            m_selection.Items.ItemSetChangedEvent += new ItemSet.ItemSetChangedHandler(OnSelectionChanged);
        }


        public Selection SelectedItems {
            get { return m_selection; }
        }

        public VectorToolSet Tools {
            get { return m_tools; }
        }
        public void SelectTool(string toolName) {
            m_tools.SelectTool(toolName);
        }

        virtual protected void OnSelectionChanged(object o, ItemSet.ItemSetChangedArgs e) {
            NotifyElementChanged();
        }


        /*
         * Item management (use these rather than directly changing .Items to get undo support)
         */
        public void AddItem(VectorStrokeItem vi) {
            Items.Add(vi);
            m_undoStack.PushAction(new AddItemAction(vi, this));

            NotifyCanvasContentsModified();
        }





        /*
         * Erasing support
         */
        ReplaceItemAction m_curEraseAction = null;

        virtual public void BeginErasing() {
            m_curEraseAction = new ReplaceItemAction(this);
        }

        virtual public void Erase( Blob eraserBlob ) {
            bool bErased = false;
            for (int i = 0; i < Items.Count; ++i) {
                VectorGraphicsItem vi = Items[i] as VectorGraphicsItem;
                if (vi == null)
                    continue;

                // save current state for undo 
                if (m_curEraseAction != null && vi.Bounds.Intersects(eraserBlob.Bounds))
                    m_curEraseAction.AddItem( vi );

                bool bErasedItem = vi.Erase( eraserBlob );
                if (bErasedItem) {
                    if (!m_vPendingOptimize.Contains( vi ))
                        m_vPendingOptimize.Add( vi );
                    bErased = true;
                }

            }
            if (bErased)
                NotifyElementChanged();
        }


        virtual public void DoneErasing() {
            OptimizeItems( m_curEraseAction );
            m_undoStack.PushAction(m_curEraseAction);

            NotifyCanvasContentsModified();
        }



        //! call after erasing to split up erased items (strokes, etc)
        void OptimizeItems( ReplaceItemAction changesAction ) {
            Dictionary<VectorGraphicsItem, List<VectorGraphicsItem> > newItems = 
                new Dictionary<VectorGraphicsItem,List<VectorGraphicsItem>>();

            for (int i = 0; i < Items.Count; ++i) {
                VectorGraphicsItem vi = Items[i] as VectorGraphicsItem;
                if ( vi == null )
                    continue;

                if (!m_vPendingOptimize.Contains( vi ))
                    continue;

                List<VectorGraphicsItem> vNew = vi.Optimize();
                newItems[vi] = vNew;

                if ( changesAction != null ) {
                    if ( vNew != null && vNew.Count > 0 )
                        changesAction.SetNewItems(vi, vNew);
                    else
                        changesAction.SetIndex(vi, Items.FindIndex(vi));
                }
            }

            foreach (VectorGraphicsItem vi in newItems.Keys) {
                List<VectorGraphicsItem> vNew = newItems[vi];

                if (vNew == null || vNew.Count == 0) {
                    Items.Remove(vi);
                    vi.Disconnect();
                } else {
                    List<Item> vNewItems = new List<Item>(vNew.Count);
                    foreach (VectorGraphicsItem vinew in vNew)
                        vNewItems.Add(vinew);
                    Items.Replace(vi, vNewItems);
                    vi.Disconnect();    // discard old item 
                }
            }

            m_vPendingOptimize.Clear();
        }




        /*
         * misc utility stuff
         */


        public Image GetCroppedImage() {
            Image img = RenderEngine.Get.GrabScreenshot(ChildAreaBounds);
            img.Crop(ColorRGBA.WHITE, 5, true);
            return img;
        }


        /*
         * Serialization support
         */


        virtual public bool Store( XmlDocument doc, XmlElement canvas_node ) {
            for (int i = 0; i < Items.Count; ++i) {
                VectorGraphicsItem vi = Items[i] as VectorGraphicsItem;
                if (vi != null) {
                    XmlElement item_node = doc.CreateElement(vi.TypeName);
                    if (vi.Store(doc, item_node))
                        canvas_node.AppendChild(item_node);
                }
            }

            return true;
        }

        virtual public bool Restore( XmlDocument doc, XmlElement canvas_node ) {
            // [RMS] not efficient!!
            VectorStrokeItem s_stroke_item = new VectorStrokeItem( null, null, null, 1, ColorRGBA.BLACK );


            for (int i = 0; i < canvas_node.ChildNodes.Count; ++i) {
                XmlElement item_node = canvas_node.ChildNodes[i] as XmlElement;
                if (item_node == null)
                    continue;

                if (item_node.Name == s_stroke_item.TypeName) {
                    VectorStrokeItem vi = new VectorStrokeItem( null, null, null, 1.0f, ColorRGBA.BLACK );
                    if (vi.Restore(doc, item_node))
                        //Items.Add( vi );
                        AddItem(vi);            // support history for loaded strokes
                }

            }


            return true;
        }





        /*
         * canvas edited notification
         */
        protected override void OnItemSetChanged( object o, ItemSet.ItemSetChangedArgs e ) {
            base.OnItemSetChanged( o, e );
            NotifyCanvasContentsModified();
        }

        public delegate void CanvasContentsModifiedHandler( VectorCanvas canvas );
        public event CanvasContentsModifiedHandler CanvasContentsModifiedEvent;
        protected void NotifyCanvasContentsModified( ) {
            if (CanvasContentsModifiedEvent != null)
                CanvasContentsModifiedEvent( this );
        } 






        /*
         *  Undo/Redo actions
         */



        virtual public void Undo( ) {
            m_undoStack.Undo();
        }


        virtual public void Redo( ) {
            m_undoStack.Redo();
        }


        class AddItemAction : UndoStack.UndoableAction 
        {
            VectorGraphicsItem m_item;
            VectorCanvas m_canvas;

            public AddItemAction(VectorGraphicsItem item, VectorCanvas canvas) {
                m_item = item;
                m_canvas = canvas;
            }

            public bool Undo() {
                m_canvas.Items.Remove(m_item);
                m_item.Disconnect();
                return true;
            }

            public bool Redo() {
                m_item.Reconnect();
                m_canvas.Items.Add(m_item);
                return true;
            }
        }



        class ReplaceItemAction : UndoStack.UndoableAction 
        {
            VectorCanvas m_canvas;
            class ReplacedItemInfo {
                public VectorGraphicsItem original;
                public List<VectorGraphicsItem> new_items;
                public int nIndex;
            };
            Dictionary<VectorGraphicsItem, ReplacedItemInfo> m_items;

            public ReplaceItemAction(VectorCanvas canvas) {
                m_canvas = canvas;
                m_items = new Dictionary<VectorGraphicsItem,ReplacedItemInfo>();
            }

            public void AddItem( VectorGraphicsItem item ) {
                if ( ! m_items.ContainsKey(item) ) {
                    ReplacedItemInfo info = new ReplacedItemInfo();
                    info.original = item.Copy();
                    info.new_items = new List<VectorGraphicsItem>();
                    info.nIndex = -1;
                    m_items.Add( item, info );
                }
            }

            public void SetNewItems(VectorGraphicsItem item, List<VectorGraphicsItem> newItems) {
                foreach (VectorGraphicsItem vi in newItems)
                    m_items[item].new_items.Add(vi);
            }

            public void SetIndex(VectorGraphicsItem item, int nIndex) {
                if (nIndex == -1)
                    System.Diagnostics.Debugger.Break();
                m_items[item].nIndex = nIndex;
            }

            public bool Undo() {

                foreach (VectorGraphicsItem vi in m_items.Keys) {
                    ReplacedItemInfo info = m_items[vi];

                    vi.Set(info.original);

                    if (info.new_items.Count > 0) {
                        List<Item> vRemoveItems = new List<Item>(info.new_items.Count);
                        foreach (VectorGraphicsItem viRemove in info.new_items)
                            vRemoveItems.Add(viRemove);
                        if (!m_canvas.Items.Replace(vRemoveItems, vi))
                            System.Diagnostics.Debugger.Break();

                        foreach (VectorGraphicsItem viRemove in info.new_items)
                            viRemove.Disconnect();
                    } else {
                        m_canvas.Items.Add(vi, info.nIndex);
                    }
                }

                return true;
            }

            public bool Redo() {

                foreach (VectorGraphicsItem vi in m_items.Keys) {
                    ReplacedItemInfo info = m_items[vi];

                    if (info.new_items.Count > 0) {
                        List<VectorGraphicsItem> vNewVectorItems = new List<VectorGraphicsItem>(info.new_items.Count);
                        List<Item> vAddItems = new List<Item>(vNewVectorItems.Count);
                        foreach (VectorGraphicsItem viAdd in info.new_items) {
                            viAdd.Reconnect();
                            vNewVectorItems.Add(viAdd);
                            vAddItems.Add(viAdd);
                        }

                        if (!m_canvas.Items.Replace(vi, vAddItems))
                            System.Diagnostics.Debugger.Break();
                    } else {
                        m_canvas.Items.Remove(vi);
                    }

                    vi.Disconnect();
                }

                return true;
            }

        }


    }
}
