using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using Tao.OpenGl;

namespace Sketcher
{
    class VectorStrokeItem : VectorGraphicsItem
    {
        class StrokeInfo {
            public List<Vector2f> vVertices;
            public List<float> vSizes;
            public List<float> vOpacities;
            public StrokeInfo( ) {
                vVertices = new List<Vector2f>();
                vSizes = new List<float>();
                vOpacities = new List<float>();
            }
            public StrokeInfo(int nCapacity) {
                vVertices = new List<Vector2f>(nCapacity);
                vSizes = new List<float>( nCapacity );
                vOpacities = new List<float>( nCapacity );
            }
            public StrokeInfo( StrokeInfo copy ) {
                vVertices = new List<Vector2f>( copy.vVertices );
                vSizes = new List<float>( copy.vSizes );
                vOpacities = new List<float>( copy.vOpacities );
            }
        }

        StrokeInfo m_stroke;
        Parameter<float> m_Width;

        StrokeInfo m_renderStroke;
        int m_renderStrokeListID = 0;

        public VectorStrokeItem(List<Vector2f> vVertices, List<float> vSizes, List<float> vOpacities, float fWidthScale, ColorRGBA color) {
            ForegroundColor = color;
            m_Width = new Parameter<float>( fWidthScale );

            m_stroke = new StrokeInfo();
            if (vVertices != null)
                m_stroke.vVertices = new List<Vector2f>( vVertices );
            if (vSizes != null)
                m_stroke.vSizes = new List<float>( vSizes );
            if (vOpacities != null)
                m_stroke.vOpacities = new List<float>( vOpacities );

            m_renderStroke = null;

            UpdateBounds();
        }

        //! copy subset of vertices in copy  (copies up to nSubEnd-1)
        public VectorStrokeItem( VectorStrokeItem copy, int nSubStart, int nSubEnd ) {
            ForegroundColor = copy.ForegroundColor;
            m_Width = new Parameter<float>( copy.m_Width.Value );

            int nCount = nSubEnd - nSubStart;
            m_stroke = new StrokeInfo( nCount );

            for (int i = nSubStart; i < nSubEnd; ++i) {
                m_stroke.vVertices.Add( copy.m_stroke.vVertices[i] );
                m_stroke.vSizes.Add( copy.m_stroke.vSizes[i] );
                m_stroke.vOpacities.Add( copy.m_stroke.vOpacities[i] );
            }

            m_renderStroke = null;

            UpdateBounds();
        }



        override public string TypeName {
            get { return "VectorStroke"; }
        }


        override public VectorGraphicsItem Copy() {
            return new VectorStrokeItem( this, 0, m_stroke.vVertices.Count );
        }
        override public bool Set(VectorGraphicsItem to) {
            VectorStrokeItem vi = to as VectorStrokeItem;
            if (vi == null)
                return false;
            ForegroundColor = vi.ForegroundColor;
            m_Width = new Parameter<float>(vi.m_Width.Value);
            m_stroke = new StrokeInfo( vi.m_stroke );
            m_renderStroke = null;
            UpdateBounds();
            return true;
        }




        //! returns true if any vertices of stroke were affected by eraser
        override public bool Erase( Blob eraserBlob ) {
            if (!eraserBlob.Bounds.Intersects( Bounds ))
                return false;

            bool bHit = false;
            for (int i = 0; i < m_stroke.vVertices.Count; ++i) {
                float fValue = eraserBlob.Value( m_stroke.vVertices[i] );
                if (fValue > 0) {
                    m_stroke.vOpacities[i] *= (1-fValue);
                    m_renderStroke = null;
                    bHit = true;
                }
            }
            return bHit;
        }


        //! split into strokes with continuous non-zero opacity
        override public List<VectorGraphicsItem> Optimize( ) {
            const float fOpacityThresh = 0.01f;

            List<VectorGraphicsItem> vNewItems = new List<VectorGraphicsItem>();

            int nSpanStart = -1;
            int iCur = 0;
            bool bInSpan = false;
            while (iCur < m_stroke.vVertices.Count) {

                if (m_stroke.vOpacities[iCur] > fOpacityThresh) {
                    if (! bInSpan) {
                        nSpanStart = iCur;
                        bInSpan = true;
                    }
                } else {
                    if (bInSpan) {
                        // create new stroke
                        vNewItems.Add( new VectorStrokeItem(
                            this, Math.Max( nSpanStart-1, 0 ), Math.Min( iCur+1, m_stroke.vVertices.Count ) ) );
                        nSpanStart = -1;
                        bInSpan = false;
                    }
                }

                ++iCur;
            }
            
            if ( bInSpan )
                vNewItems.Add( new VectorStrokeItem(
                    this, Math.Max( nSpanStart-1, 0 ), m_stroke.vVertices.Count ) );

            return vNewItems; 
        }



        StrokeInfo Resample( float fMinSampleDist, bool bSubSample ) {
            StrokeInfo resampled = new StrokeInfo();

            int i = 0;
            float fAccumDist = 0.0f;
            while (i < m_stroke.vVertices.Count) {
                bool bAdd = false;
                if (i == 0 || i == m_stroke.vVertices.Count-1)
                    bAdd = true;
                else {
                    float fDist = m_stroke.vVertices[i].Dist( m_stroke.vVertices[i-1] );
                    fAccumDist += fDist;
                    if (fAccumDist > fMinSampleDist)
                        bAdd = true;
                }
                if (bAdd) {
                    resampled.vVertices.Add( m_stroke.vVertices[i] );
                    resampled.vOpacities.Add( m_stroke.vOpacities[i] );
                    resampled.vSizes.Add( m_stroke.vSizes[i] );
                    fAccumDist = 0.0f;
                }
                ++i;
            }

            return resampled;
        }



        //! draw object
        override public void Render() {
            VectorCanvas canvas = (VectorCanvas)Parent;
            bool bSelected = canvas.SelectedItems.IsSelected(this);

            //RenderEngine.Get.DrawBoxOutline( Bounds, ColorRGBA.RED );

            ValidateRenderStroke();
            Gl.glCallList(m_renderStrokeListID);
            //RenderEngine.Get.DrawStroke(m_renderStroke.vVertices, m_renderStroke.vSizes, m_renderStroke.vOpacities, m_Width.Value, Opacity,
            //        (bSelected) ? ColorRGBA.RED : ForegroundColor);
            //RenderEngine.Get.DrawStroke(m_stroke.vVertices, m_stroke.vSizes, m_stroke.vOpacities, m_Width.Value, Opacity,
            //        (bSelected) ? ColorRGBA.RED : ForegroundColor);
               
        }


        void UpdateBounds( ) {
            if (m_stroke.vVertices.Count > 0) {
                Box2f bounds = new Box2f( m_stroke.vVertices[0], m_stroke.vVertices[0] );
                for (int i = 1; i < m_stroke.vVertices.Count; ++i)
                    bounds.Contain( m_stroke.vVertices[i] );
                bounds.Expand( m_Width.Value );
                Resize( bounds );
            }
        }



        void ValidateRenderStroke( ) {
            if (m_renderStroke != null)
                return;

            m_renderStroke = Resample( 2.0f, false );
            //StrokeInfo useStroke = m_renderStroke;
            StrokeInfo useStroke = m_stroke;

            if (m_renderStrokeListID == 0)
                m_renderStrokeListID = Gl.glGenLists( 1 );

            // [TODO] support display lists in render engine (somehow)
            Gl.glNewList( m_renderStrokeListID, Gl.GL_COMPILE );
            RenderEngine.Get.DrawStroke(useStroke.vVertices, useStroke.vSizes, useStroke.vOpacities, 
                m_Width.Value, 1.0f, ForegroundColor );
            Gl.glEndList();
        }



        /*
         * VectorGraphicsItem overrides
         */

        override public float IsHit(Vector2f vClick) {
            // disabled for now...
            return 0;
        }


        override public void NotifySizeChanged() {
            // ignore for now (cannot change size...)
        }

        override public bool Store( XmlDocument doc, XmlElement item_node ) {

            item_node.SetAttribute("Width", m_Width.Value.ToString());
            item_node.SetAttribute( "VertexCount", m_stroke.vVertices.Count.ToString() );

            StringBuilder builder = new StringBuilder();

            foreach (Vector2f v in m_stroke.vVertices) {
                builder.Append( v.X ); builder.Append( " " );
                builder.Append( v.Y ); builder.Append( " " );
            }
            XmlElement vertices = doc.CreateElement( "Vertices" );
            vertices.AppendChild( doc.CreateTextNode(builder.ToString()) );
            item_node.AppendChild( vertices );

            builder = new StringBuilder();
            foreach (float f in m_stroke.vSizes) {
                builder.Append( f ); builder.Append( " " );
            }
            XmlElement sizes = doc.CreateElement( "Sizes" );
            sizes.AppendChild( doc.CreateTextNode(builder.ToString()) );
            item_node.AppendChild( sizes );

            builder = new StringBuilder();
            foreach (float f in m_stroke.vOpacities) {
                builder.Append( f ); builder.Append( " " );
            }
            XmlElement opacities = doc.CreateElement( "Opacities" );
            opacities.AppendChild( doc.CreateTextNode( builder.ToString() ) );
            item_node.AppendChild( opacities );
            
            return true;
        }

        override public bool Restore( XmlDocument doc, XmlElement item_node ) {

            m_stroke = new StrokeInfo();

            m_Width.SetNewValue( 1.0f );
            if (item_node.HasAttribute( "Width" ))
                m_Width.SetNewValue( float.Parse( item_node.GetAttribute( "Width" ) ) );

            int nVertexCount = 0;
            if (item_node.HasAttribute( "VertexCount" ))
                nVertexCount = int.Parse( item_node.GetAttribute( "VertexCount" ) );

            XmlNodeList vVertices = item_node.GetElementsByTagName( "Vertices" );
            if (vVertices.Count != 1)
                return false;
            XmlElement vertices_node = vVertices[0] as XmlElement;
            string sVerts = vertices_node.FirstChild.Value;
            string[] sCoords = sVerts.Split( ' ' );
            for ( int i = 0; i < nVertexCount; ++i ) {
                float fX = float.Parse( sCoords[2*i] );
                float fY = float.Parse( sCoords[2*i+1] );
                m_stroke.vVertices.Add( new Vector2f( fX, fY ) );
            }

            XmlNodeList vSizes = item_node.GetElementsByTagName( "Sizes" );
            if (vSizes.Count != 1)
                return false;
            XmlElement sizes_node = vSizes[0] as XmlElement;
            string[] sSizes = sizes_node.FirstChild.Value.Split( ' ' );
            for (int i = 0; i < nVertexCount; ++i) {
                float fSize = float.Parse( sSizes[i] );
                m_stroke.vSizes.Add( fSize );
            }

            XmlNodeList vOpacities = item_node.GetElementsByTagName( "Opacities" );
            if (vOpacities.Count != 1)
                return false;
            XmlElement opacities_node = vOpacities[0] as XmlElement;
            string[] sOpacities = opacities_node.FirstChild.Value.Split( ' ' );
            for (int i = 0; i < nVertexCount; ++i) {
                float fOpacity = float.Parse( sOpacities[i] );
                m_stroke.vOpacities.Add( fOpacity );
            }

            m_renderStroke = null;
            UpdateBounds();

            return true;
        }
    }
}
